/* ====================================================================
 * The Apache Software License, Version 1.1
 *
 * Copyright (c) 2000-2003 The Apache Software Foundation.  All rights
 * reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 *    if any, must include the following acknowledgment:
 *       "This product includes software developed by the
 *        Apache Software Foundation (http://apache.org/)."
 *    Alternately, this acknowledgment may appear in the software itself,
 *    if and wherever such third-party acknowledgments normally appear.
 *
 * 4. The names "Apache" and "Apache Software Foundation", "Tapestry" 
 *    must not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact apache@apache.org.
 *
 * 5. Products derived from this software may not be called "Apache" 
 *    or "Tapestry", nor may "Apache" or "Tapestry" appear in their 
 *    name, without prior written permission of the Apache Software Foundation.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE TAPESTRY CONTRIBUTOR COMMUNITY
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * ====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of the Apache Software Foundation.  For more
 * information on the Apache Software Foundation, please see
 * <http://www.apache.org/>.
 *
 */

package org.apache.tapestry.util.io;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.OutputStream;
import java.io.Serializable;
import java.util.zip.GZIPInputStream;
import java.util.zip.GZIPOutputStream;

import org.apache.tapestry.Tapestry;

/**
 *  The most complicated of the adaptors, this one takes an arbitrary serializable
 *  object, serializes it to binary, and encodes it in a Base64 encoding.
 *
 *  <p>Encoding and decoding of Base64 strings uses code adapted from work in the public
 *  domain originally written by Jonathan Knudsen and published in
 *  O'reilly's "Java Cryptography". Note that we use a <em>modified</em> form of Base64 encoding,
 *  with URL-safe characters to encode the 62 and 63 values and the pad character.
 *
 *  <p>TBD:  Work out some class loader issues involved in deserializing.
 *
 *  @author Howard Lewis Ship
 *  @version $Id: SerializableAdaptor.java,v 1.2 2003/05/28 13:41:47 hlship Exp $
 * 
 **/

class SerializableAdaptor implements ISqueezeAdaptor
{
    private static final String PREFIX = "O";

    /**
     *  The PAD character, appended to the end of the string to make things
     *  line up.  In normal Base64, this is the character '='.
     *
     **/

    private static final char PAD = '.';

    /**
     *  Representation for the 6-bit code 63, normally '+' in Base64.
     *
     **/

    private static final char CH_62 = '-';

    /**
     *  Representation for the 6-bit code 64, normally '/' in Base64.
     *
     **/

    private static final char CH_63 = '_';

    public String squeeze(DataSqueezer squeezer, Object data) throws IOException
    {
        ByteArrayOutputStream bos = null;
        GZIPOutputStream gos = null;
        ObjectOutputStream oos = null;
        byte[] byteData = null;

        try
        {
            bos = new ByteArrayOutputStream();
            gos = new GZIPOutputStream(bos);
            oos = new ObjectOutputStream(gos);

            oos.writeObject(data);
            oos.close();
        }
        finally
        {
            close(oos);
            close(gos);
            close(bos);
        }

        byteData = bos.toByteArray();

        StringBuffer encoded = new StringBuffer(2 * byteData.length);
        char[] base64 = new char[4];

        encoded.append(PREFIX);

        for (int i = 0; i < byteData.length; i += 3)
        {
            encodeBlock(byteData, i, base64);
            encoded.append(base64);
        }

        return encoded.toString();
    }

    private void close(OutputStream stream)
    {
        if (stream != null)
        {
            try
            {
                stream.close();
            }
            catch (IOException ex)
            {
                // Ignore.
            }
        }
    }

    private void close(InputStream stream)
    {
        if (stream != null)
        {
            try
            {
                stream.close();
            }
            catch (IOException ex)
            {
                // Ignore.
            }
        }
    }
    public Object unsqueeze(DataSqueezer squeezer, String string) throws IOException
    {
        ByteArrayInputStream bis = null;
        GZIPInputStream gis = null;
        ObjectInputStream ois = null;
        byte[] byteData;

        // Strip off the first character and decode the rest.

        byteData = decode(string.substring(1));

        try
        {
            bis = new ByteArrayInputStream(byteData);
            gis = new GZIPInputStream(bis);
            ois = new ResolvingObjectInputStream(squeezer.getResolver(), gis);

            return ois.readObject();
        }
        catch (ClassNotFoundException ex)
        {
            // The message is the name of the class.

            throw new IOException(
                Tapestry.format("SerializableAdaptor.class-not-found", ex.getMessage()));
        }
        finally
        {
            close(ois);
            close(gis);
            close(bis);
        }
    }

    public void register(DataSqueezer squeezer)
    {
        squeezer.register(PREFIX, Serializable.class, this);
    }

    private static void encodeBlock(byte[] raw, int offset, char[] base64) throws IOException
    {
        int block = 0;
        int slack = raw.length - offset - 1;
        int end = (slack >= 2) ? 2 : slack;

        for (int i = 0; i <= end; i++)
        {
            byte b = raw[offset + i];
            int neuter = (b < 0) ? b + 256 : b;
            block += neuter << (8 * (2 - i));
        }

        for (int i = 0; i < 4; i++)
        {
            int sixbit = (block >>> (6 * (3 - i))) & 0x3f;
            base64[i] = getChar(sixbit);
        }

        if (slack < 1)
            base64[2] = PAD;

        if (slack < 2)
            base64[3] = PAD;
    }

    protected static char getChar(int sixBit) throws IOException
    {
        if (sixBit >= 0 && sixBit <= 25)
            return (char) ('A' + sixBit);

        if (sixBit >= 26 && sixBit <= 51)
            return (char) ('a' + (sixBit - 26));

        if (sixBit >= 52 && sixBit <= 61)
            return (char) ('0' + (sixBit - 52));

        if (sixBit == 62)
            return CH_62;

        if (sixBit == 63)
            return CH_63;

        throw new IOException(
            Tapestry.format("SerializableAdaptor.unable-to-convert", Integer.toString(sixBit)));
    }

    public static byte[] decode(String string) throws IOException
    {
        int pad = 0;
        char[] base64 = string.toCharArray();

        for (int i = base64.length - 1; base64[i] == PAD; i--)
            pad++;

        int length = base64.length * 6 / 8 - pad;
        byte[] raw = new byte[length];
        int rawIndex = 0;

        for (int i = 0; i < base64.length; i += 4)
        {
            int block =
                (getValue(base64[i]) << 18)
                    + (getValue(base64[i + 1]) << 12)
                    + (getValue(base64[i + 2]) << 6)
                    + (getValue(base64[i + 3]));

            for (int j = 0; j < 3 && rawIndex + j < raw.length; j++)
                raw[rawIndex + j] = (byte) ((block >> (8 * (2 - j))) & 0xff);

            rawIndex += 3;
        }

        return raw;
    }

    private static int getValue(char c) throws IOException
    {
        if (c >= 'A' && c <= 'Z')
            return c - 'A';

        if (c >= 'a' && c <= 'z')
            return c - 'a' + 26;

        if (c >= '0' && c <= '9')
            return c - '0' + 52;

        if (c == CH_62)
            return 62;

        if (c == CH_63)
            return 63;

        // Pad character

        if (c == PAD)
            return 0;

        throw new IOException(
            Tapestry.format(
                "SerializableAdaptor.unable-to-interpret-char",
                new String(new char[] { c })));
    }

}